// 
// Copyright (c) 2009 Boris Schaeling <boris@highscore.de> 
// 
// Distributed under the Boost Software License, Version 1.0. (See accompanying 
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt) 
// 

#ifndef BOOST_ASIO_POSIX_SIGNAL_HANDLER_IMPL_HPP 
#define BOOST_ASIO_POSIX_SIGNAL_HANDLER_IMPL_HPP 

#include <boost/thread.hpp> 
#include <map> 
#include <deque> 
#include <signal.h> 

namespace boost { 
namespace asio { 
namespace posix { 

class signal_handler_impl 
{ 
public: 
    signal_handler_impl() 
        : run_(true) 
    { 
    } 

    ~signal_handler_impl() 
    { 
        for (sigactions_t::iterator it = sigactions_.begin(); it != sigactions_.end(); ++it) 
            sigaction(it->first, &it->second, 0); 
    } 

    void add_signal(int signal, struct sigaction oldact) 
    { 
        sigactions_.insert(sigactions_t::value_type(signal, oldact)); 
    } 

    void remove_signal(int signal) 
    { 
        sigactions_t::iterator it = sigactions_.find(signal); 
        if (it != sigactions_.end()) 
        { 
            sigaction(signal, &it->second, 0); 
            sigactions_.erase(it); 
        } 
    } 

    void destroy() 
    { 
        boost::unique_lock<boost::recursive_mutex> lock(pending_signals_mutex_); 
        run_ = false; 
        pending_signals_cond_.notify_all(); 
    } 

    int wait(boost::system::error_code &ec) 
    { 
        boost::unique_lock<boost::recursive_mutex> lock(pending_signals_mutex_); 
        while (run_ && pending_signals_.empty()) 
            pending_signals_cond_.wait(lock); 
        int signal = 0; 
        if (!pending_signals_.empty()) 
        { 
            signal = pending_signals_.front(); 
            pending_signals_.pop_front(); 
        } 
        else 
            ec = boost::asio::error::operation_aborted; 
        return signal; 
    } 

    void handle(int signal) 
    { 
        // A recursive mutex must be used as this method is called by the signal handler 
        // which can run in any thread - including a thread which locked already the mutex. 
        boost::unique_lock<boost::recursive_mutex> lock(pending_signals_mutex_); 
        pending_signals_.push_back(signal); 
        pending_signals_cond_.notify_all(); 
    } 

private: 
    typedef std::map<int, struct sigaction> sigactions_t; 
    sigactions_t sigactions_; 
    boost::recursive_mutex pending_signals_mutex_; 
    boost::condition_variable_any pending_signals_cond_; 
    bool run_; 
    std::deque<int> pending_signals_; 
}; 

} 
} 
} 

#endif 
